@@ -71,7 +71,7 @@ _local_option = click.option(
 
 def _error(msg):
     es_telemetry.send_telemetry(msg, success=False)
-    click.echo(f"ERROR: {msg}", err=True)
+    click.echo(f"\nERROR: {msg}", err=True)
     sys.exit(1)
 
 
@@ -95,12 +95,8 @@ def click_exception_wrapper(command):
                 return result
             except KeyboardInterrupt:
                 _error(f"{command}: aborted by keyboard interrupt")
-            except TimedOutError as err:
-                _error(f"{command}: hit timeout error: {err}")
-            except FatalError as err:
-                _error(f"{command}: hit fatal error: {err}")
-            except StatusError as err:
-                _error(f"{command}: hit status error: {err}")
+            except (TimedOutError, FatalError, StatusError) as err:
+                _error(f"{command}: hit error: {err}")
             except Exception as err:
                 log.exception("hit unexpected error")
                 _error(f"{command}: hit unexpected error: {err}")
@@ -213,11 +209,11 @@ def _get_requester():
     return requester
 
 
-def _wait_on_namespace_resources(namespace, timeout, db_only=False):
+def _wait_on_namespace_resources(namespace, timeout, db_only=False, defer_status_errors=False):
     if db_only:
-        wait_for_db_resources(namespace, timeout)
+        wait_for_db_resources(namespace, timeout, defer_status_errors)
     else:
-        wait_for_all_resources(namespace, timeout)
+        wait_for_all_resources(namespace, timeout, defer_status_errors)
 
 
 def _validate_reservation_duration(ctx, param, value):
@@ -291,7 +287,7 @@ _ns_list_options = [
     ),
 ]
 
-_timeout_option = [
+_timeout_options = [
     click.option(
         "--timeout",
         "-t",
@@ -299,7 +295,13 @@ _timeout_option = [
         type=int,
         default=600,
         help="timeout in sec (default = 600) to wait for resources to be ready",
-    )
+    ),
+    click.option(
+        "--defer-status-errors",
+        is_flag=True,
+        default=False,
+        help="do not exit immediately if status errors seen (e.g. ImagePullBackOff). Default: false",
+    ),
 ]
 
 
@@ -401,8 +403,7 @@ def _app_or_component_selector(ctx, param, this_value):
     # validate that opposing options are not both set to 'all'
     if this_value.select_all and other_value.select_all:
         raise click.BadParameter(
-            f"'all' cannot be specified on both this option"
-            f" and its opposite '{other_param_name}'"
+            f"'all' cannot be specified on both this option and its opposite '{other_param_name}'"
         )
 
     # validate that the same app was not used in opposing options
@@ -515,8 +516,7 @@ _process_options = _app_source_options + [
         "--set-parameter",
         "-p",
         help=(
-            "Override parameter for a component using format "
-            "'<component>/<parameter name>=<value>'"
+            "Override parameter for a component using format '<component>/<parameter name>=<value>'"
         ),
         multiple=True,
         callback=_validate_set_parameter,
@@ -545,9 +545,9 @@ _process_options = _app_source_options + [
     click.option(
         "--remove-resources",
         help=(
-            "Remove resource limits and requests on ClowdApp configs "
-            "for specific components or apps. Prefix the app name with "
-            "'app:', otherwise specify the component name. (default: all)"
+            "Remove untrusted (defined in README) resource limits/requests on "
+            "ClowdApp/ClowdJob/CJI objects for specific components or apps. Prefix the app name "
+            "with 'app:', otherwise specify the component name. (default: 'all')"
         ),
         type=str,
         multiple=True,
@@ -556,9 +556,9 @@ _process_options = _app_source_options + [
     click.option(
         "--no-remove-resources",
         help=(
-            "Don't remove resource limits and requests on ClowdApp configs "
-            "for specific components or apps. Prefix the app name with "
-            "'app:', otherwise specify the component name. (default: none)"
+            "Preserve resource limits/requests even if untrusted (defined in README) on "
+            "ClowdApp/ClowdJob/CJI objects for specific components or apps. Prefix the app name "
+            "with 'app:', otherwise specify the component name. (default: none)"
         ),
         type=str,
         multiple=True,
@@ -599,6 +599,11 @@ _process_options = _app_source_options + [
         type=str,
         multiple=True,
     ),
+    click.option(
+        "--exclude-components",
+        type=str,
+        help="Comma-separated list of components to exclude from deployment",
+    ),
     click.option(
         "--frontends",
         "-F",
@@ -824,9 +829,11 @@ def _list_namespaces(available, mine, output):
 
 @namespace.command("reserve")
 @options(_ns_reserve_options)
-@options(_timeout_option)
+@options(_timeout_options)
 @click_exception_wrapper("namespace reserve")
-def _cmd_namespace_reserve(name, requester, duration, pool, timeout, local, force):
+def _cmd_namespace_reserve(
+    name, requester, duration, pool, timeout, local, force, defer_status_errors
+):
     """Reserve an ephemeral namespace"""
     ns = _check_and_reserve_namespace(name, requester, duration, pool, timeout, local, force)
     click.echo(ns.name)
@@ -891,16 +898,19 @@ def _cmd_namespace_extend(namespace, duration, local):
     default=False,
     help="Only wait for DB resources owned by ClowdApps to be ready",
 )
-@options(_timeout_option)
-def _cmd_namespace_wait_on_resources(namespace, timeout, db_only):
+@options(_timeout_options)
+def _cmd_namespace_wait_on_resources(namespace, timeout, db_only, defer_status_errors):
     """Wait for rolled out resources to be ready in namespace"""
     if not namespace:
         namespace = current_namespace_or_error()
     try:
-        _wait_on_namespace_resources(namespace, timeout, db_only=db_only)
+        _wait_on_namespace_resources(namespace, timeout, db_only, defer_status_errors)
     except TimedOutError as err:
-        log.error("Hit timeout error: %s", err)
+        log.error("hit timeout error: %s", err)
         _error("namespace wait timed out")
+    except StatusError as err:
+        log.error("hit status error: %s", err)
+        _error("namespace resources have status errors")
 
 
 @namespace.command("describe")
@@ -951,6 +961,10 @@ def _get_apps_config(
         log.info("fetching apps config using source: %s", source)
         apps_config = get_appsfile_apps(config)
 
+    # handle git ref/image substitutions if reference environment was provided
+    if ref_env:
+        apps_config = sub_refs(apps_config, ref_env, fallback_ref_env, preferred_params)
+
     # merge remote apps config with local app config
     local_apps = get_local_apps(config)
     apps_config = merge_app_configs(apps_config, local_apps, local_config_method)
@@ -967,10 +981,6 @@ def _get_apps_config(
                 # re-raise with a bit more context
                 raise FatalError(f"{str(err)}, hit on app {app_name}")
 
-    # handle git ref/image substitutions if reference environment was provided
-    if ref_env:
-        apps_config = sub_refs(apps_config, ref_env, fallback_ref_env, preferred_params)
-
     return apps_config
 
 
@@ -1023,6 +1033,8 @@ def _process(
     local,
     frontends,
     preferred_params,
+    namespace,
+    exclude_components,
 ):
     apps_config = _get_apps_config(
         source,
@@ -1051,6 +1063,8 @@ def _process(
         component_filter,
         local,
         frontends,
+        namespace,
+        exclude_components,
     )
     return processor.process()
 
@@ -1105,6 +1119,7 @@ def _cmd_process(
     local,
     frontends,
     preferred_params,
+    exclude_components,
 ):
     """Fetch and process application templates"""
     clowd_env = _get_env_name(namespace, clowd_env)
@@ -1132,6 +1147,8 @@ def _cmd_process(
         local,
         frontends,
         preferred_params,
+        namespace,
+        exclude_components,
     )
     print(json.dumps(processed_templates, indent=2))
 
@@ -1228,6 +1245,30 @@ def _check_and_reserve_namespace(name, requester, duration, pool, timeout, local
     return reserve_namespace(name, requester, duration, pool, timeout, local)
 
 
+def _deploy_err_handler(err, no_release_on_fail, reserved_new_ns, reserve, ns):
+    if isinstance(err, KeyboardInterrupt):
+        msg = "keyboard interrupt"
+    else:
+        msg = "deploy failed"
+
+    if str(err):
+        msg += f": {str(err)}"
+
+    if isinstance(err, (KeyboardInterrupt, TimedOutError, FatalError, StatusError)):
+        log.error(msg)
+    else:
+        log.exception("hit unexpected error!")
+
+    try:
+        if not no_release_on_fail and reserved_new_ns and not reserve:
+            # if we auto-reserved this ns, auto-release it on failure unless
+            # --no-release-on-fail was requested
+            log.info("releasing namespace '%s'", ns)
+            release_reservation(namespace=ns)
+    finally:
+        _error(msg)
+
+
 @main.command("deploy")
 @options(_process_options)
 @click.option(
@@ -1277,7 +1318,7 @@ def _check_and_reserve_namespace(name, requester, duration, pool, timeout, local
     help="Do not release namespace reservation if deployment fails",
 )
 @options(_ns_reserve_options)
-@options(_timeout_option)
+@options(_timeout_options)
 def _cmd_config_deploy(
     app_names,
     source,
@@ -1304,6 +1345,7 @@ def _cmd_config_deploy(
     duration,
     timeout,
     no_release_on_fail,
+    exclude_components,
     component_filter,
     import_secrets,
     import_configmaps,
@@ -1314,6 +1356,7 @@ def _cmd_config_deploy(
     pool,
     force,
     preferred_params,
+    defer_status_errors,
 ):
     """Process app templates and deploy them to a cluster"""
     if not has_clowder():
@@ -1346,19 +1389,6 @@ def _cmd_config_deploy(
 
     clowd_env = _get_env_name(ns, clowd_env)
 
-    def _err_handler(err):
-        try:
-            if not no_release_on_fail and reserved_new_ns and not reserve:
-                # if we auto-reserved this ns, auto-release it on failure unless
-                # --no-release-on-fail was requested
-                log.info("releasing namespace '%s'", ns)
-                release_reservation(namespace=ns)
-        finally:
-            msg = "deploy failed"
-            if str(err):
-                msg += f": {str(err)}"
-            _error(msg)
-
     try:
         log.info("processing app templates...")
         apps_config = _process(
@@ -1384,6 +1414,8 @@ def _cmd_config_deploy(
             local,
             frontends,
             preferred_params,
+            namespace,
+            exclude_components,
         )
         log.debug("app configs:\n%s", json.dumps(apps_config, indent=2))
         if not apps_config["items"]:
@@ -1392,22 +1424,9 @@ def _cmd_config_deploy(
             log.info("applying app configs...")
             apply_config(ns, apps_config)
             log.info("waiting on resources for max of %dsec...", timeout)
-            _wait_on_namespace_resources(ns, timeout)
-    except KeyboardInterrupt as err:
-        log.error("aborted by keyboard interrupt!")
-        _err_handler(err)
-    except TimedOutError as err:
-        log.error("hit timeout error: %s", err)
-        _err_handler(err)
-    except FatalError as err:
-        log.error("hit fatal error: %s", err)
-        _err_handler(err)
-    except StatusError as err:
-        log.error("hit status error: %s", err)
-        _err_handler(err)
-    except Exception as err:
-        log.exception("hit unexpected error!")
-        _err_handler(err)
+            _wait_on_namespace_resources(ns, timeout, False, defer_status_errors)
+    except (KeyboardInterrupt, Exception) as err:
+        _deploy_err_handler(err, no_release_on_fail, reserved_new_ns, reserve, ns)
     else:
         log.info("successfully deployed to namespace %s", ns)
         es_telemetry.send_telemetry("successful deployment")
@@ -1416,9 +1435,9 @@ def _cmd_config_deploy(
             ns_url = f"{url}/k8s/cluster/projects/{ns}"
             log.info("namespace url: %s", ns_url)
             log.info(
-                "resource usage dashboard for namespace '%s': https://grafana.app-sre.devshift.net/d/jRY7KLnVz?var-namespace=%s",
-                ns,
+                "resource usage dashboard for namespace '%s': %s",
                 ns,
+                conf.RESOURCE_DASHBOARD_URL.format(namespace=ns),
             )
         click.echo(ns)
 
@@ -1454,7 +1473,7 @@ def _cmd_process_clowdenv(namespace, quay_user, clowd_env, template_file, local)
 @click.option(
     "--secrets-dir",
     type=str,
-    help=("Import secrets from this directory (default: " "$XDG_CONFIG_HOME/bonfire/secrets/)"),
+    help=("Import secrets from this directory (default: $XDG_CONFIG_HOME/bonfire/secrets/)"),
     default=conf.DEFAULT_SECRETS_DIR,
 )
 @click.option(
@@ -1468,7 +1487,7 @@ def _cmd_process_clowdenv(namespace, quay_user, clowd_env, template_file, local)
     default=conf.DEFAULT_CONFIGMAPS_DIR,
 )
 @options(_ns_reserve_options)
-@options(_timeout_option)
+@options(_timeout_options)
 @click_exception_wrapper("deploy-env")
 def _cmd_deploy_clowdenv(
     namespace,
@@ -1486,6 +1505,7 @@ def _cmd_deploy_clowdenv(
     local,
     pool,
     force,
+    defer_status_errors,
 ):
     """Process ClowdEnv template and deploy to a cluster"""
     if not has_clowder():
@@ -1510,7 +1530,7 @@ def _cmd_deploy_clowdenv(
         namespace = wait_for_clowd_env_target_ns(clowd_env)
 
     log.info("waiting on resources for max of %dsec...", timeout)
-    _wait_on_namespace_resources(namespace, timeout)
+    _wait_on_namespace_resources(namespace, timeout, False, defer_status_errors)
 
     clowd_env_name = find_clowd_env_for_ns(namespace)["metadata"]["name"]
 
@@ -1570,7 +1590,7 @@ def _cmd_process_iqe_cji(
 @click.option("--namespace", "-n", help="Namespace to deploy to", type=str, required=True)
 @options(_iqe_cji_process_options)
 @options(_ns_reserve_options)
-@options(_timeout_option)
+@options(_timeout_options)
 @click_exception_wrapper("deploy-iqe-cji")
 def _cmd_deploy_iqe_cji(
     namespace,
@@ -1599,6 +1619,7 @@ def _cmd_deploy_iqe_cji(
     custom_env_vars,
     pool,
     force,
+    defer_status_errors,
 ):
     """Process IQE CJI template, apply it, and wait for it to start running."""
     if not has_clowder():
@@ -1638,7 +1659,7 @@ def _cmd_deploy_iqe_cji(
     apply_config(namespace, cji_config)
 
     log.info("waiting on CJI '%s' for max of %dsec...", cji_name, timeout)
-    pod_name = wait_on_cji(namespace, cji_name, timeout)
+    pod_name = wait_on_cji(namespace, cji_name, timeout, defer_status_errors)
     log.info("pod '%s' related to CJI '%s' in ns '%s' is running", pod_name, cji_name, namespace)
     click.echo(pod_name)
 
@@ -1720,7 +1741,7 @@ def _cmd_apps_what_depends_on(
 def main_with_handler():
     try:
         main()
-    except FatalError as err:
+    except (StatusError, FatalError) as err:
         _error(str(err))
 
 
