@@ -3,6 +3,7 @@ import json
 import logging
 import re
 import traceback
+from typing import Optional
 import uuid
 from pathlib import Path
 
@@ -30,23 +31,115 @@ def _process_template(*args, **kwargs):
     return processed_template
 
 
-def _remove_resource_config(items):
-    for i in items:
-        if i["kind"] != "ClowdApp":
+def _get_trusted_config(data: dict, key1: str, key2: str, path: str, component_params: dict):
+    """
+    Check for presence of a trusted param being used for the value.
+
+    A trusted parameter is defined as:
+    1. matching the expected regex pattern
+    2. the parameter value is set in the component's deploy config
+
+    Returns True if we determine this is a trusted value
+    """
+    regex = conf.TRUSTED_RESOURCE_REGEX.get(key1, {}).get(key2)
+
+    if not regex or not isinstance(regex, str):
+        raise ValueError("not able to find conf.TRUSTED_RESOURCE_REGEX[{key1}][{key2}]")
+
+    value = data.get(key1, {}).get(key2)
+    if value:
+        match = False
+        in_params = False
+
+        match = re.match(regex, str(value))
+        if match and match.groups()[0] in component_params:
+            in_params = True
+
+        log.debug(
+            "value '%s', regex r'%s', matches=%s, in params=%s",
+            value,
+            regex,
+            bool(match),
+            in_params,
+        )
+
+        if match and in_params:
+            return value
+        else:
+            log.debug("deleted untrusted config at '%s.%s.%s'", path, key1, key2)
+            return None
+
+
+def _remove_untrusted_configs(
+    data,
+    params,
+    path="",
+    current_dict=None,
+):
+    """
+    Locate configurations within 'data' and remove them if not trusted.
+
+    Checks to see if resource configurations match the regex specified in
+    conf.TRUSTED_RESOURCE_REGEX
+
+    Also checks that if a request is defined, a corresponding limit is defined
+    for the same resource and vice-versa.
+    """
+    if path.endswith(".resources"):
+        resources = current_dict["resources"]
+        cpu_request = _get_trusted_config(resources, "requests", "cpu", path, params)
+        cpu_limit = _get_trusted_config(resources, "limits", "cpu", path, params)
+        mem_request = _get_trusted_config(resources, "requests", "memory", path, params)
+        mem_limit = _get_trusted_config(resources, "limits", "memory", path, params)
+
+        if any([cpu_request, cpu_limit]) and not all([cpu_request, cpu_limit]):
+            log.debug("'%s' cpu config needs both request and limit, removing cpu config", path)
+            cpu_request = None
+            cpu_limit = None
+        if any([mem_request, mem_limit]) and not all([mem_request, mem_limit]):
+            log.debug("'%s' mem config needs both request and limit, removing mem config", path)
+            mem_request = None
+            mem_limit = None
+
+        # if val is null, omit it from final dict
+        if not cpu_request:
+            del resources["requests"]["cpu"]
+        if not cpu_limit:
+            del resources["limits"]["cpu"]
+        if not mem_request:
+            del resources["requests"]["memory"]
+        if not mem_limit:
+            del resources["limits"]["memory"]
+
+    elif isinstance(data, dict):
+        for key, value in copy.copy(data).items():
+            _remove_untrusted_configs(value, params, path + f".{key}", current_dict=data)
+
+    elif isinstance(data, list):
+        for index, value in enumerate(copy.copy(data)):
+            _remove_untrusted_configs(value, params, path + f"[{index}]")
+
+
+def _remove_untrusted_configs_for_template(template, params):
+    """
+    Removes untrusted configurations within the template.
+
+        A configuration is trusted if:
+        1. the value is defined using a template parameter
+        2. the parameter follows the proper syntax convention
+        3. the parameter is defined on the component in its deployment config
+
+    Checks template to see if any resources of kind listed in 'config.TRUSTED_CHECK_KINDS'
+    are found. If so, analyzes the config on those resources to see if any need to be removed.
+    """
+    for obj in template.get("objects", []):
+        kind = obj.get("kind")
+        if kind not in conf.TRUSTED_CHECK_KINDS:
             continue
 
-        removed = False
-        for d in i["spec"].get("deployments", []):
-            if "resources" in d["podSpec"]:
-                del d["podSpec"]["resources"]
-                removed = True
-        for p in i["spec"].get("pods", []):
-            if "resources" in p:
-                del p["resources"]
-                removed = True
-
-        if removed:
-            log.debug("removed resources from ClowdApp '%s'", i["metadata"]["name"])
+        name = obj.get("metadata", {}).get("name")
+        log.debug("checking resources on %s '%s'", kind, name)
+        _remove_untrusted_configs(obj, params)
 
 
 def _remove_dependency_config(items):
@@ -112,6 +205,8 @@ def process_clowd_env(target_ns, quay_user, env_name, template_path, local=True)
 
     params = dict()
     params["ENV_NAME"] = env_name
+    # TODO: revisit need for below param once FEO has a more developed config management system
+    params["FRONTEND_CONTEXT_NAME"] = env_name
     if quay_user:
         quay_user = quay_user.replace("_", "-")
         params["PULL_SECRET_NAME"] = f"{quay_user}-pull-secret"
@@ -293,6 +388,23 @@ class TemplateProcessor:
                 parsed_app_names.add(entry)
         return parsed_app_names
 
+    @staticmethod
+    def _parse_exclude_components(exclude_components):
+        """Parse exclude_components which can be None, a string, or a list."""
+        if exclude_components is None:
+            return None
+
+        if isinstance(exclude_components, str):
+            # Handle comma-separated string
+            return [
+                component.strip()
+                for component in exclude_components.split(",")
+                if component.strip()
+            ]
+
+        # Assume it's already a list
+        return list(exclude_components)
+
     @staticmethod
     def _find_dupe_components(components_for_app):
         """Make sure no component is listed more than once across all apps."""
@@ -464,6 +576,8 @@ class TemplateProcessor:
         component_filter,
         local,
         frontends,
+        namespace=None,
+        exclude_components=None,
     ):
         self.apps_config = apps_config
         self.requested_app_names = self._parse_app_names(app_names)
@@ -481,6 +595,8 @@ class TemplateProcessor:
         self.component_filter = component_filter
         self.local = local
         self.frontends = frontends
+        self.namespace = namespace
+        self.exclude_components = self._parse_exclude_components(exclude_components)
 
         self._validate()
 
@@ -558,13 +674,20 @@ class TemplateProcessor:
             rf = RepoFile.from_config(component)
             # override template ref if requested
             self._sub_ref(component_name, rf)
+            log.debug(
+                "component: '%s' fetching template using git ref '%s'", component_name, rf.ref
+            )
             commit, template_content = rf.fetch()
         except Exception as err:
             log.error("failed to fetch template file for %s", component_name)
             log.debug(traceback.format_exc())
             raise FatalError(err)
 
-        template = yaml.safe_load(template_content)
+        try:
+            template = yaml.safe_load(template_content)
+        except Exception as err:
+            log.exception("failed to parse template content to yaml for %s", component_name)
+            raise FatalError(err)
 
         # fetch component parameters
         params = component.get("parameters", {})
@@ -573,22 +696,23 @@ class TemplateProcessor:
         if "IMAGE_TAG" not in params:
             params["IMAGE_TAG"] = commit[:7]
 
+        # set NAMESPACE on this component only if it is current unset
+        if "NAMESPACE" not in params:
+            params["NAMESPACE"] = self.namespace
+
         # always override ENV_NAME
         params["ENV_NAME"] = self.clowd_env
+        # TODO: revisit need for below param once FEO has a more developed config management system
+        params["FRONTEND_CONTEXT_NAME"] = self.clowd_env
 
         # override other specific parameters on this component if requested by user at runtime
         self._sub_params(component_name, params)
         log.debug("parameters for component '%s': %s", component_name, params)
 
-        new_items = _process_template(template, params, self.local)["items"]
-
-        # override the tags for all occurences of an image if requested
-        new_items = self._sub_image_tags(new_items)
-
         # evaluate --remove-resources/--no-remove-resources
         app_name = self._get_app_for_component(component_name)
 
-        # if app/component is trusted, do not remove its resources
+        # if entire app or component is marked trusted, do not remove any resources
         should_remove_resources = False
         if app_name in conf.TRUSTED_APPS:
             log.debug("should_remove: app '%s' listed in trusted apps", app_name)
@@ -604,7 +728,13 @@ class TemplateProcessor:
             )
         log.debug("should_remove_resources evaluates to %s", should_remove_resources)
         if should_remove_resources:
-            _remove_resource_config(new_items)
+            _remove_untrusted_configs_for_template(template, params)
+
+        # process the template
+        new_items = _process_template(template, params, self.local)["items"]
+
+        # override the tags for all occurences of an image if requested
+        new_items = self._sub_image_tags(new_items)
 
         # evaluate --remove-dependencies/--no-remove-dependencies
         should_remove_deps = _should_remove(
@@ -694,7 +824,7 @@ class TemplateProcessor:
             processed_component.optional_deps_handled = True
 
         for component_name in all_dependencies:
-            self._process_component(component_name, app_name, in_recursion=True)
+            self._process_component(component_name, app_name, in_recursion=True, is_dependency=True)
 
     def _handle_dependencies(self, app_name, processed_component, in_recursion):
         items = processed_component.items
@@ -702,10 +832,10 @@ class TemplateProcessor:
             for name in conf.AUTO_ADDED_FRONTEND_DEPENDENCIES:
                 if name not in self.processed_components:
                     log.info("auto-adding %s as dependency for frontend resource", name)
-                    self._process_component(name, app_name, in_recursion)
+                    self._process_component(name, app_name, in_recursion, is_dependency=True)
         self._add_dependencies_to_config(app_name, processed_component, in_recursion)
 
-    def _process_component(self, component_name, app_name, in_recursion):
+    def _process_component(self, component_name, app_name, in_recursion, is_dependency=False):
         if component_name in self.processed_components:
             log.debug("template already processed for component '%s'", component_name)
             processed_component = self.processed_components[component_name]
@@ -721,9 +851,13 @@ class TemplateProcessor:
                 )
                 items = []
 
-            self.k8s_list["items"].extend(items)
             processed_component = ProcessedComponent(component_name, items)
             self.processed_components[component_name] = processed_component
+            reason = self._component_skip_check(component_name, is_dependency)
+            if reason:
+                log.info("skipping component '%s', %s", component_name, reason)
+            else:
+                self.k8s_list["items"].extend(items)
 
         if self.get_dependencies:
             # recursively process to add config for dependent apps to self.k8s_list
@@ -735,12 +869,28 @@ class TemplateProcessor:
         for component in app_cfg["components"]:
             component_name = component["name"]
             log.debug("app '%s' has component '%s'", app_name, component_name)
-            if self.component_filter and component_name not in self.component_filter:
-                log.debug(
-                    "skipping component '%s', not found in --component filter", component_name
-                )
-                continue
-            self._process_component(component_name, app_name, in_recursion=False)
+            self._process_component(
+                component_name, app_name, in_recursion=False, is_dependency=False
+            )
+
+    def _component_skip_check(self, component_name, is_dependency) -> Optional[str]:
+        skip_reasons = [
+            (
+                not is_dependency
+                and self.component_filter
+                and component_name not in self.component_filter,
+                "not found in --component filter",
+            ),
+            (
+                self.exclude_components and component_name in self.exclude_components,
+                "found in --exclude-components list",
+            ),
+        ]
+
+        for condition, reason in skip_reasons:
+            if condition:
+                return reason
+        return None
 
     def process(self, app_names=None):
         if not app_names:
